################################################################################
# Copyright (c) 2014-2022, National Research Foundation (SARAO)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

"""Tests for :mod:`katsdpsigproc.rfi.twodflag`."""

import concurrent.futures

import numpy as np
import pytest
import scipy.interpolate
from scipy.ndimage import gaussian_filter1d, gaussian_filter

from katsdpsigproc.rfi import twodflag


class TestAsbool:
    def _test(self, dtype, expect_view):
        a = np.array([0, 1, 1, 0, 1, 0, 0, 1], dtype)
        expected = a.astype(np.bool_)
        out = twodflag._asbool(a)
        assert out.dtype == np.bool_
        np.testing.assert_array_equal(expected, out)
        if expect_view:
            # Change a, out must change because it is a view
            a[0] = not a[0]
            assert out[0] == bool(a[0])

    def test_uint8(self):
        self._test(np.uint8, True)

    def test_uint16(self):
        self._test(np.uint16, False)

    def test_bool(self):
        self._test(np.bool_, True)


class TestAverageFreq:
    def setup_method(self):
        self.small_data = np.arange(30, dtype=np.float32).reshape(5, 6, 1).repeat(2, axis=2)
        self.small_flags = np.zeros(self.small_data.shape, np.bool_)
        self.small_flags[3, :, 0] = 1
        self.small_flags[:, 4, 0] = 1
        self.small_flags[2, 0, :] = 1
        self.small_flags[2, 5, :] = 1

    def test_one(self):
        """_average_freq with 1 channel must have no effect on unflagged data."""
        avg_data, avg_flags = twodflag._average_freq(self.small_data, self.small_flags,
                                                     twodflag._as_min_dtype(1))
        expected = self.small_data.copy()
        expected[self.small_flags] = 0
        assert avg_data.dtype == np.float32
        assert avg_flags.dtype == np.bool_
        np.testing.assert_array_equal(np.moveaxis(expected, -1, 0), avg_data)
        np.testing.assert_array_equal(np.moveaxis(self.small_flags, -1, 0), avg_flags)

    def test_divides(self):
        """Test _average_freq when averaging factor divides in exactly."""
        expected_data = np.array([
            [
                [0.5, 2.5, 5.0],
                [6.5, 8.5, 11.0],
                [13.0, 14.5, 0.0],
                [0.0, 0.0, 0.0],
                [24.5, 26.5, 29.0]
            ],
            [
                [0.5, 2.5, 4.5],
                [6.5, 8.5, 10.5],
                [13.0, 14.5, 16.0],
                [18.5, 20.5, 22.5],
                [24.5, 26.5, 28.5]
            ]], np.float32)
        expected_flags = np.array([
            [
                [False, False, False],
                [False, False, False],
                [False, False, True],
                [True, True, True],
                [False, False, False]
            ],
            [[False, False, False]] * 5])
        avg_data, avg_flags = twodflag._average_freq(self.small_data, self.small_flags,
                                                     twodflag._as_min_dtype(2))
        assert avg_data.dtype == np.float32
        assert avg_flags.dtype == np.bool_
        np.testing.assert_array_equal(expected_data, avg_data)
        np.testing.assert_array_equal(expected_flags, avg_flags)

    def test_uneven(self):
        """Test _average_freq when averaging factor does not divide number of channels."""
        expected_data = np.array([
            [
                [1.5, 5.0],
                [7.5, 11.0],
                [14.0, 0.0],
                [0.0, 0.0],
                [25.5, 29.0],
            ],
            [
                [1.5, 4.5],
                [7.5, 10.5],
                [14.0, 16.0],
                [19.5, 22.5],
                [25.5, 28.5]
            ]], np.float32)
        expected_flags = np.array([
            [
                [False, False],
                [False, False],
                [False, True],
                [True, True],
                [False, False]
            ], [[False, False]] * 5], np.bool_)
        avg_data, avg_flags = twodflag._average_freq(self.small_data, self.small_flags,
                                                     twodflag._as_min_dtype(4))
        assert avg_data.dtype == np.float32
        assert avg_flags.dtype == np.bool_
        np.testing.assert_array_equal(expected_data, avg_data)
        np.testing.assert_array_equal(expected_flags, avg_flags)


def test_time_median():
    """Test for :func:`katsdpsigproc.rfi.twodflag._time_median`."""
    data = np.array([
        [2.0, 1.0, 2.0, 5.0],
        [3.0, 1.0, 8.0, 6.0],
        [4.0, 1.0, 4.0, 7.0],
        [5.0, 1.0, 5.0, 6.5],
        [1.5, 1.0, 1.5, 5.5]], np.float32)
    flags = np.array([
        [0, 1, 0, 1],
        [0, 1, 1, 0],
        [0, 1, 0, 1],
        [0, 1, 0, 1],
        [0, 1, 0, 1]], np.bool_)
    out_data, out_flags = twodflag._time_median(data, flags)
    expected_data = np.array([[3.0, 0.0, 3.0, 6.0]], np.float32)
    expected_flags = np.array([[0, 1, 0, 0]], np.bool_)
    np.testing.assert_array_equal(expected_data, out_data)
    np.testing.assert_array_equal(expected_flags, out_flags)


class TestMedianAbs:
    """Test :func:`.twodflag._median_abs` and :func:`.twodflag._median_abs_axis0`."""

    def setup_method(self):
        self.data = np.array([[-2.0, -6.0, 4.5], [1.5, 3.3, 0.5]], np.float32)
        self.flags = np.array([[0, 0, 0], [0, 1, 0]], np.uint8)

    def test(self):
        out = twodflag._median_abs(self.data, self.flags)
        assert out == 2.0

    def test_all_flagged(self):
        out = twodflag._median_abs(self.data, np.ones_like(self.flags))
        assert np.isnan(out)

    def test_axis0(self):
        out = twodflag._median_abs_axis0(self.data, self.flags)
        expected = np.array([[1.75, 6.0, 2.5]])
        np.testing.assert_array_equal(expected, out)

    def test_axis0_all_flagged(self):
        self.flags[:, 1] = True
        out = twodflag._median_abs_axis0(self.data, self.flags)
        expected = np.array([[1.75, np.nan, 2.5]])
        np.testing.assert_array_equal(expected, out)


class TestLinearlyInterpolateNans:
    """Tests for :func:`katsdpsigproc.rfi.twodflag._linearly_interpolate_nans`."""

    def setup_method(self):
        self.y = np.array([np.nan, np.nan, 4.0, np.nan, np.nan, 10.0, np.nan, -2.0, np.nan, np.nan])
        self.expected = np.array([4.0, 4.0, 4.0, 6.0, 8.0, 10.0, 4.0, -2.0, -2.0, -2.0])

    def test_basic(self):
        twodflag._linearly_interpolate_nans1d(self.y)
        np.testing.assert_allclose(self.expected, self.y)

    def test_no_nans(self):
        y = self.expected[:]
        twodflag._linearly_interpolate_nans1d(y)
        np.testing.assert_allclose(self.expected, y)

    def test_all_nans(self):
        self.y[:] = np.nan
        self.expected[:] = 0
        twodflag._linearly_interpolate_nans1d(self.y)
        np.testing.assert_array_equal(self.expected, self.y)

    def test_float32(self):
        expected = self.expected.astype(np.float32)
        y = self.y.astype(np.float32)
        twodflag._linearly_interpolate_nans1d(y)
        np.testing.assert_allclose(expected, y, rtol=1e-6)

    def test_2d(self):
        y = np.zeros((3, self.y.size))
        y[0, :] = self.y
        y[1, :] = self.expected
        y[2, :] = np.nan
        expected = np.zeros_like(y)
        expected[0, :] = self.expected
        expected[1, :] = self.expected
        expected[2, :] = 0
        twodflag._linearly_interpolate_nans(y)
        np.testing.assert_allclose(expected, y)


class TestBoxGaussianFilter:
    def test_one_pass(self):
        """Test that _box_gaussian_filter1d places the box correctly."""
        a = np.array([50.0, 10.0, 60.0, -70.0, 30.0, 20.0, -15.0], np.float32)
        b = np.empty_like(a)
        twodflag._box_gaussian_filter1d(a, 2, b, 1)
        np.testing.assert_equal(
            np.array([24.0, 10.0, 16.0, 10.0, 5.0, -7.0, 7.0], np.float32), b)

    def test_width(self):
        """Impulse response must have approximately correct standard deviation, \
        and must be symmetric with sum 1."""
        a = np.zeros((1, 200), np.float32)
        a[:, a.size // 2] = 1.0
        sigma = np.array([0.0, 10.0])
        b = np.empty_like(a)
        twodflag._box_gaussian_filter(a, sigma, b)
        x = np.arange(a.size) - a.size // 2
        total = np.sum(b)
        np.testing.assert_allclose(1.0, total, rtol=1e-5)
        mean = np.sum(x * b)
        np.testing.assert_allclose(0.0, mean, atol=1e-5)
        std = np.sqrt(np.sum(x * x * b))
        # Very loose test, because box_gaussian_filter1d quantises
        np.testing.assert_allclose(std, sigma[1], atol=1)

    def test_bad_sigma_dim(self):
        a = np.zeros((50, 50), np.float32)
        with pytest.raises(ValueError):
            twodflag._box_gaussian_filter(a, np.array([3.0]), a)

    def test_2d(self):
        rs = np.random.RandomState(seed=1)
        shape = (77, 53)
        sigma = np.array([8, 2.3])
        data = rs.uniform(size=shape).astype(np.float32)
        expected = gaussian_filter(data, sigma, mode='constant')
        actual = np.zeros_like(data)
        twodflag._box_gaussian_filter(data, sigma, actual)
        np.testing.assert_allclose(expected, actual, rtol=1e-1)

    def test_axes(self):
        """Test that the axes are handled consistently."""
        rs = np.random.RandomState(seed=1)
        shape = (77, 53)
        data = rs.uniform(size=shape).astype(np.float32)
        out0 = np.zeros_like(data)
        out1 = np.zeros_like(data)
        twodflag._box_gaussian_filter(data, np.array([8.0, 0.0]), out0)
        twodflag._box_gaussian_filter(data.T, np.array([0.0, 8.0]), out1.T)
        np.testing.assert_array_equal(out0, out1)

    def test_edge(self):
        """Test that values outside the boundary are handled like zeros."""
        rs = np.random.RandomState(seed=1)
        data = np.zeros((1, 200), np.float32)
        core = data[:, 80:120]
        core[:] = rs.uniform(size=core.shape)
        fdata = np.ones_like(data)
        fcore = np.ones_like(core)
        twodflag._box_gaussian_filter(data, np.array([0.0, 3.0]), fdata)
        twodflag._box_gaussian_filter(core, np.array([0.0, 3.0]), fcore)
        np.testing.assert_allclose(fdata[:, 80:120], fcore, rtol=1e-5)


class TestMaskedGaussianFilter:
    def setup_method(self):
        self.rs = np.random.RandomState(seed=1)
        shape = (77, 53)
        self.data = self.rs.uniform(size=shape).astype(np.float32)
        self.flags = self.rs.uniform(size=shape) >= 0.5

    def _get_expected(self, sigma, truncate):
        weight = 1.0 - self.flags
        data = self.data * weight
        for i, (s, t) in enumerate(zip(sigma, truncate)):
            weight = gaussian_filter1d(weight, s, axis=i, mode='constant', truncate=t)
            data = gaussian_filter1d(data, s, axis=i, mode='constant', truncate=t)
        with np.errstate(invalid='ignore'):
            data /= weight
        return data

    def test_basic(self):
        sigma = np.array([5, 2.3])
        expected = self._get_expected(sigma, (4.0, 4.0))
        actual = np.ones_like(expected)
        twodflag.masked_gaussian_filter(self.data, self.flags, sigma, actual)
        np.testing.assert_allclose(expected, actual, rtol=1e-1)

    def test_nan(self):
        # Set a big block of zeros to get NaNs in the result
        self.flags[:] = False
        self.flags[30:70, 10:40] = True
        # To match NaN positions, we need to match the footprint of the kernels
        sigma = np.array([3, 3.3])
        passes = 4
        radius = [int(0.5 * np.sqrt(12.0 * s**2 / passes + 1)) for s in sigma]
        truncate = [passes * r / s for (r, s) in zip(radius, sigma)]
        expected = self._get_expected(sigma, truncate)
        actual = np.ones_like(self.data)
        twodflag.masked_gaussian_filter(self.data, self.flags, sigma, actual)
        np.testing.assert_allclose(expected, actual, rtol=1e-1)
        # Check that some NaNs were generated
        assert np.sum(np.isnan(expected)) > 0


class TestGetBackground2D:
    """Tests for :func:`katsdpsigproc.rfi.twodflag._get_background2d`.

    This is a difficult function to test, because it's not really practical to
    determine expected results by hand. The tests mainly check corner cases
    where large regions are flagged.
    """

    def setup_method(self):
        self.shape = (95, 86)
        self.data = np.ones(self.shape, np.float32) * 7.5
        self.flags = np.zeros(self.shape, np.uint8)

    def _get_background2d(self, data, flags=None, iterations=1, spike_width=(10.0, 10.0),
                          reject_threshold=2.0, freq_chunks=None):
        if flags is None:
            flags = np.zeros(data.shape, np.uint8)
        if freq_chunks is None:
            freq_chunks = np.array([0, data.shape[1]])
        spike_width = np.array(spike_width, np.float32)
        return twodflag._get_background2d(data, flags, iterations, spike_width, reject_threshold,
                                          freq_chunks)

    def test_no_flags(self):
        background = self._get_background2d(self.data)
        assert background.dtype == np.float32
        # It's all constant, so background and output should match.
        # It won't be exact though, because the Gaussian filter accumulates
        # errors as it sums.
        np.testing.assert_allclose(self.data, background, rtol=1e-5)

    def test_all_flagged(self):
        self.flags[:] = True
        background = self._get_background2d(self.data, self.flags)
        assert background.dtype == np.float32
        np.testing.assert_array_equal(np.zeros(self.shape, np.float32), background)

    def test_in_flags(self):
        # This needs to be done carefully, because getbackground_2d does
        # internal masking on outliers too. We give every 3rd time a higher
        # power and flag it.
        self.data[::3] = 20.0
        self.flags[::3] = True
        background = self._get_background2d(self.data, self.flags)
        expected = np.ones_like(self.data) * 7.5
        np.testing.assert_allclose(expected, background, rtol=1e-5)

    def test_interpolate(self):
        """Test linear interpolation across completely flagged data."""
        # Block of channels is 7.5, then a block is flagged (still 7.5), then
        # a block is 3.0.
        self.data[:, 70:] = 3.0
        self.flags[:, 30:70] = True
        # The setup above has no deviation from the background, which makes the
        # outlier rejection unstable, so we add noise to half the timesteps, and test
        # them at lower precision. We use uniform noise to guarantee no outliers.
        rs = np.random.RandomState(seed=1)
        self.data[:50, :] += rs.uniform(-0.001, 0.001, self.data[0:50, :].shape)

        # The rejection threshold is adjusted, because the default doesn't do
        # well when only about half the data is noisy.
        background = self._get_background2d(self.data, self.flags, spike_width=(2.5, 2.5),
                                            reject_threshold=5.0)
        expected = np.zeros_like(self.data)
        expected[:, :37] = 7.5
        expected[:, 63:] = 3.0
        expected[:, 37:63] = np.linspace(7.5, 3.0, 26)
        np.testing.assert_allclose(expected[56:], background[56:], rtol=1e-4)
        np.testing.assert_allclose(expected[:56], background[:56], rtol=1e-2)

    def test_iterations(self):
        expected = self.data.copy()
        # Add some noise
        rs = np.random.RandomState(seed=1)
        self.data += rs.standard_normal(self.data.shape) * 0.1
        # Add a "spike" that's larger than the initial spike_width to check
        # that it gets masked out.
        self.data[20:50, 30:80] += 15

        background = self._get_background2d(self.data, iterations=3)
        np.testing.assert_allclose(expected, background, rtol=1e-2)


class TestSumThreshold:
    def setup_method(self):
        self.small_data = np.arange(30, dtype=np.float32).reshape(5, 6)
        self.small_flags = np.zeros(self.small_data.shape, np.bool_)
        self.small_flags[3, :] = 1
        self.small_flags[:, 4] = 1
        self.small_flags[2, 0] = 1
        self.small_flags[2, 5] = 1
        self.outlier_nsigma = 4.5
        self.rho = 1.3
        self.windows = np.array([1, 2, 4, 8])

    def test_sum_threshold_all_flagged(self):
        self.small_flags[:] = True
        out_flags = twodflag._sum_threshold(self.small_data, self.small_flags, 0,
                                            np.array([1, 2, 4]), self.outlier_nsigma, self.rho)
        np.testing.assert_array_equal(np.zeros_like(self.small_flags), out_flags)

    def _test_sum_threshold_basic(self, axis):
        rs = np.random.RandomState(seed=1)
        data = rs.standard_normal((100, 90)).astype(np.float32) * 3.0
        rfi = np.zeros_like(data)
        # Add some individual spikes and some bad channels
        rfi[10, 20] = 100.0
        rfi[80, 80] = -100.0
        rfi[:, 40] = rs.uniform(80.0, 120.0, size=(100,))
        rfi[:, 2] = -rfi[:, 40]
        # Smaller but wider spike
        rfi[:, 60:67] = rs.uniform(15.0, 20.0, size=(100, 7))
        rfi[:, 10:17] = -rfi[:, 60:67]
        in_flags = np.zeros(data.shape, np.bool_)
        expected_flags = rfi != 0
        data += rfi
        if axis == 0:
            # Swap axes around so that we're doing essentially the same test
            rfi = rfi.T.copy()
            data = data.T.copy()
            in_flags = in_flags.T.copy()
        out_flags = twodflag._sum_threshold(data, in_flags, axis,
                                            self.windows, self.outlier_nsigma, self.rho)
        if axis == 0:
            out_flags = out_flags.T
        # Due to random data, won't get perfect agreement, but should get close
        errors = np.sum(expected_flags != out_flags)
        assert errors / data.size < 0.01
        # Check for exact match on the individual spikes
        for region in (np.s_[8:13, 18:23], np.s_[78:83, 78:83]):
            np.testing.assert_equal(expected_flags[region], out_flags[region])

    def test_sum_threshold_time(self):
        self._test_sum_threshold_basic(axis=0)

    def test_sum_threshold_frequency(self):
        self._test_sum_threshold_basic(axis=1)

    def test_sum_threshold_existing(self):
        rs = np.random.RandomState(seed=1)
        data = rs.standard_normal((100, 90)).astype(np.float32) * 3.0
        in_flags = np.zeros(data.shape, np.bool_)
        # Corrupt but pre-flag just under half the data, which will skew the
        # noise estimate if not taken into account.
        data[:48] += 1000.0
        in_flags[:48] = True
        # Add some spikes that should be just under the detection limit.
        data[70, 0] = 12.5
        data[70, 1] = -12.5
        # Add some spikes that should still be detected.
        data[70, 2] = 20.0
        data[70, 3] = -20.0
        # Test it
        out_flags = twodflag._sum_threshold(data, in_flags, 0, self.windows, 5, self.rho)
        np.testing.assert_array_equal([False, False, True, True], out_flags[70, :4])


class TestSumThresholdFlagger:
    """Tests for :class:`katsdpsigproc.rfi.twodflag.SumThresholdFlagger`."""

    def setup_method(self):
        self.flagger = twodflag.SumThresholdFlagger()

    def _make_background(self, shape, rs):
        """Simulate a bandpass with some smooth variation."""
        ntime, nfreq, n_bl = shape
        nx = 10
        x = np.linspace(0.0, nfreq, nx)
        y = np.ones((ntime, nx, n_bl)) * 2.34
        y[:, 0, :] = 0.1
        y[:, -1, :] = 0.1
        y[:] += rs.uniform(0.0, 0.1, y.shape)
        f = scipy.interpolate.interp1d(x, y, axis=1, kind='cubic', assume_sorted=True)
        return f(np.arange(nfreq))

    def _make_data(self, flagger, rs, shape=(234, 345, 1)):
        background = self._make_background(shape, rs).astype(np.float32)
        data = background + (rs.standard_normal(shape) * 0.1).astype(np.float32)
        rfi = np.zeros(shape, np.float32)
        # Some completely bad channels and bad times
        rfi[12, :] = 1
        rfi[20:25, :] = 1
        rfi[:, 17] = 1
        rfi[:, 200:220] = 1
        # Some mostly bad channels and times
        rfi[30, :300] = 1
        rfi[50:, 80] = 1
        # Some smaller blocks of RFI
        rfi[60:65, 100:170] = 1
        rfi[150:200, 150:153] = 1
        expected = rfi.astype(np.bool_)
        # The mostly-bad channels and times must be fully flagged
        expected[30, :] = True
        expected[:, 80] = True
        data += rfi * rs.standard_normal(shape) * 3.0
        # Channel that is slightly biased, but wouldn't be picked up in a single dump
        data[:, 260] += 0.2 * flagger.average_freq
        expected[:, 260] = True
        # Test input NaN value flagged on output
        data[225, 225] = np.nan
        expected[225, 225] = True
        in_flags = np.zeros(shape, np.bool_)
        # Pre-flag some channels, and make those values NaN (because cal
        # currently does this - but should be fixed).
        in_flags[:, 185:190] = True
        data[:, 185:190] = np.nan
        return np.abs(data), in_flags, expected

    def _test_get_flags(self, flagger):
        rs = np.random.RandomState(seed=1)
        data, in_flags, expected = self._make_data(flagger, rs)

        orig_data = data.copy()
        orig_in_flags = in_flags.copy()
        out_flags = flagger.get_flags(data, in_flags)
        # Check that the original values aren't disturbed
        np.testing.assert_equal(orig_data, data)
        np.testing.assert_equal(orig_in_flags, in_flags)

        # Check the results. Everything that's expected must be flagged,
        # everything within time_extend and freq_extend kernels may be
        # flagged. A small number of values outside this may be flagged.
        # The backgrounding doesn't currently handle the edges of the band
        # well, so for now we also allow the edges to be flagged too.
        # TODO: improve _get_background2d so that it better fits a slope
        # at the edges of the passband.
        allowed = expected | in_flags
        allowed[:-1] |= allowed[1:]
        allowed[1:] |= allowed[:-1]
        allowed[:, :-1] |= allowed[:, 1:]
        allowed[:, 1:] |= allowed[:, :-1]
        allowed[:, :40] = True
        allowed[:, -40:] = True
        missing = expected & ~out_flags
        extra = out_flags & ~allowed
        # Uncomment for debugging failures
        # import matplotlib.pyplot as plt
        # plt.imshow(expected[..., 0] + 2 * out_flags[..., 0] + 4 * allowed[..., 0])
        # plt.show()
        assert missing.sum() == 0
        assert extra.sum() / data.size < 0.03

    def test_get_flags(self):
        self._test_get_flags(self.flagger)

    def test_get_flags_single_chunk(self):
        flagger = twodflag.SumThresholdFlagger(freq_chunks=1)
        self._test_get_flags(flagger)

    def test_get_flags_many_chunks(self):
        # Number of chunks can't be too high, otherwise the block of channels
        # affected by RFI isn't detected as it mostly falls into one chunk.
        flagger = twodflag.SumThresholdFlagger(freq_chunks=15)
        self._test_get_flags(flagger)

    def test_get_flags_average_freq(self):
        flagger = twodflag.SumThresholdFlagger(average_freq=2)
        self._test_get_flags(flagger)

    # TODO: fix up the overflagging of the background in the flagger,
    # which currently causes this to fail.
    @pytest.mark.skip(reason='Backgrounder overflags edges of the slope')
    def test_get_flags_iterations(self):
        flagger = twodflag.SumThresholdFlagger(background_iterations=3)
        self._test_get_flags(flagger)

    def _test_get_flags_all_flagged(self, flagger):
        data = np.zeros((100, 80, 4), np.float32)
        in_flags = np.ones(data.shape, np.bool_)
        out_flags = flagger.get_flags(data, in_flags)
        np.testing.assert_array_equal(np.zeros_like(in_flags), out_flags)

    def test_get_flags_all_flagged(self):
        self._test_get_flags_all_flagged(self.flagger)

    def test_get_flags_all_flagged_average_freq(self):
        flagger = twodflag.SumThresholdFlagger(average_freq=4)
        self._test_get_flags_all_flagged(flagger)

    def test_variable_noise(self):
        """Noise level that varies across the band."""
        rs = np.random.RandomState(seed=1)
        shape = (234, 345, 1)
        # For this test we use a flat background, to avoid the issues with
        # bandpass estimation in sloped regions.
        background = np.ones(shape, np.float32) * 11
        # Noise level that varies from 0 to 1 across the band
        noise = rs.standard_normal(shape)
        noise *= np.arange(shape[1])[np.newaxis, :, np.newaxis] / shape[1]
        noise = noise.astype(np.float32)
        noise[100, 17] = 1.0    # About 20 sigma - must be detected
        noise[200, 170] = 1.0   # About 2 sigma -  must not be detected
        data = np.abs(background + noise)
        in_flags = np.zeros(shape, np.bool_)
        out_flags = self.flagger.get_flags(data, in_flags)
        assert out_flags[100, 17, 0]
        assert not out_flags[200, 170, 0]

    def _test_parallel(self, pool):
        """Test that parallel execution gets same results as serial."""
        rs = np.random.RandomState(seed=1)
        data, in_flags, expected = self._make_data(self.flagger, rs, shape=(234, 512, 32))
        out_serial = self.flagger.get_flags(data, in_flags)
        out_parallel = self.flagger.get_flags(data, in_flags, pool=pool)
        np.testing.assert_array_equal(out_serial, out_parallel)

    def test_thread_pool(self):
        with concurrent.futures.ThreadPoolExecutor(4) as pool:
            self._test_parallel(pool)

    def test_process_pool(self):
        with concurrent.futures.ProcessPoolExecutor(4) as pool:
            self._test_parallel(pool)



################################################################################
# Copyright (c) 2014-2022, National Research Foundation (SARAO)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

"""Tests for RFI flagging wrappers.

The backgrounding and thresholding have separate tests - this module just tests
that they can be glued together properly.
"""

import numpy as np

from katsdpsigproc.rfi import host, device
from katsdpsigproc.abc import AbstractContext, AbstractCommandQueue
from .. import complex_normal


_vis: np.ndarray
_spikes: np.ndarray
_input_flags: np.ndarray


def setup():   # type: () -> None
    global _vis, _spikes, _input_flags
    shape = (117, 131)
    # Use a fixed seed to make the test repeatable
    rs = np.random.RandomState(seed=1)
    _vis = complex_normal(rs, size=shape)
    # Pick 1/16 of samples to be RFI
    # Give them random amplitude from a range, and random phase
    _spikes = rs.random_sample(shape) < 1.0 / 16.0
    _spikes = _spikes.astype(np.uint8)
    rfi_amp = rs.random_sample(shape) * 20.0 + 50.0
    rfi_phase = rs.random_sample(shape) * (2j * np.pi)
    rfi = rfi_amp * np.exp(rfi_phase)
    _vis += _spikes * rfi
    _vis = _vis.astype(np.complex64)
    _input_flags = (rs.random_sample(shape) < 1.0 / 16.0).astype(np.uint8) * 2


def test_flagger_host() -> None:
    global _vis, _spikes, _input_flags
    background = host.BackgroundMedianFilterHost(13)
    noise_est = host.NoiseEstMADHost()
    threshold = host.ThresholdSimpleHost(11.0)
    flagger = host.FlaggerHost(background, noise_est, threshold)
    flags = flagger(_vis)
    np.testing.assert_equal(_spikes, flags)
    # Test again, this time with channel flags
    flags = flagger(_vis, _input_flags[:, 0])
    input_flags = np.broadcast_to(_input_flags[:, 0:1], _vis.shape)
    expected = np.where(input_flags, 0, _spikes)
    np.testing.assert_equal(expected, flags)
    # And again with full input flags
    flags = flagger(_vis, _input_flags)
    expected = np.where(_input_flags, 0, _spikes)
    np.testing.assert_equal(expected, flags)


def check_flagger_device(use_flags: device.BackgroundFlags,
                         transpose_noise_est: bool, transpose_threshold: bool,
                         context: AbstractContext, command_queue: AbstractCommandQueue) -> None:
    global _vis, _spikes, _input_flags
    background = device.BackgroundMedianFilterDeviceTemplate(context, 13, use_flags=use_flags)
    if transpose_noise_est:
        noise_est = device.NoiseEstMADTDeviceTemplate(
            context, 1024)       # type: device.AbstractNoiseEstDeviceTemplate
    else:
        noise_est = device.NoiseEstMADDeviceTemplate(context, tuning={'wgsx': 8, 'wgsy': 8})
    threshold = device.ThresholdSimpleDeviceTemplate(
        context, transpose_threshold, tuning={'wgsx': 8, 'wgsy': 8})
    flagger_device = device.FlaggerDeviceTemplate(background, noise_est, threshold)
    flagger = device.FlaggerHostFromDevice(
        flagger_device, command_queue, threshold_args=dict(n_sigma=11.0))
    if use_flags == device.BackgroundFlags.CHANNEL:
        flags = flagger(_vis, _input_flags[:, 0])
        input_flags = np.broadcast_to(_input_flags[:, 0:1], _vis.shape)
        expected = np.where(input_flags, 0, _spikes)
        np.testing.assert_equal(expected, flags)
    elif use_flags == device.BackgroundFlags.FULL:
        flags = flagger(_vis, _input_flags)
        expected = np.where(_input_flags, 0, _spikes)
        np.testing.assert_equal(expected, flags)
    else:
        flags = flagger(_vis)
        np.testing.assert_equal(_spikes, flags)


def test_flagger_device(context: AbstractContext, command_queue: AbstractCommandQueue) -> None:
    check_flagger_device(device.BackgroundFlags.NONE, False, False, context, command_queue)


def test_flagger_device_transpose_noise_est(context: AbstractContext,
                                            command_queue: AbstractCommandQueue) -> None:
    """Test device flagger with a transposed noise estimator."""
    check_flagger_device(device.BackgroundFlags.CHANNEL, True, False, context, command_queue)


def test_flagger_device_transpose_threshold(context: AbstractContext,
                                            command_queue: AbstractCommandQueue) -> None:
    """Test device flagger with a transposed thresholder."""
    check_flagger_device(device.BackgroundFlags.FULL, False, True, context, command_queue)


def test_flagger_device_transpose_both(context: AbstractContext,
                                       command_queue: AbstractCommandQueue) -> None:
    """Test device flagger with a transposed noise estimator and thresholder."""
    check_flagger_device(device.BackgroundFlags.NONE, True, True, context, command_queue)



################################################################################
# Copyright (c) 2014-2022, National Research Foundation (SARAO)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

"""Tests for :mod:`katsdpsigproc.fill`."""

from typing import cast

import numpy as np
import pytest

from katsdpsigproc import accel
from katsdpsigproc import fill
from katsdpsigproc.abc import AbstractContext, AbstractCommandQueue


class TestFill:
    @classmethod
    def pad_dimension(cls, dim: accel.Dimension, extra: int) -> None:
        """Modify `dim` to have at least `extra` padding."""
        newdim = accel.Dimension(dim.size, min_padded_size=dim.size + extra)
        newdim.link(dim)

    def test_fill(self, context: AbstractContext, command_queue: AbstractCommandQueue) -> None:
        shape = (75, 63)
        template = fill.FillTemplate(context, np.uint32, 'unsigned int')
        fn = template.instantiate(command_queue, shape)
        data_slot = cast(accel.IOSlot, fn.slots['data'])
        self.pad_dimension(data_slot.dimensions[0], 5)
        self.pad_dimension(data_slot.dimensions[1], 10)
        fn.ensure_all_bound()
        data = fn.buffer('data')
        # Do the fill
        fn.set_value(0xDEADBEEF)
        fn()
        # Check the result, including padding
        ret = data.get(command_queue)
        assert ret.base is not None
        ret = ret.base
        np.testing.assert_equal(ret, 0xDEADBEEF)

    @pytest.mark.force_autotune
    def test_autotune(self, context: AbstractContext) -> None:
        """Test that autotuner runs successfully."""
        fill.FillTemplate(context, np.uint8, 'unsigned char')



################################################################################
# Copyright (c) 2021, National Research Foundation (SARAO)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

"""Plugin for testing with pytest.

See the :ref:`testing-pytest` section of the documentation for details.
"""

import itertools
from typing import Generator

import pytest

from . import abc, accel, tune


@pytest.fixture
def patch_autotune(request, monkeypatch) -> None:
    autotuner = tune.stub_autotuner
    if request.node.get_closest_marker("force_autotune"):
        autotuner = tune.force_autotuner
    monkeypatch.setattr(tune, "autotuner_impl", autotuner)


@pytest.fixture
def context(device: abc.AbstractDevice,
            patch_autotune) -> Generator[abc.AbstractContext, None, None]:
    # Make the context current (for CUDA contexts). Ideally the test
    # should not depend on this, but PyCUDA leaks memory if objects
    # are deleted without the context current.
    with device.make_context() as context:
        yield context


@pytest.fixture
def command_queue(context: abc.AbstractContext) -> abc.AbstractCommandQueue:
    return context.create_command_queue()


def pytest_addoption(parser):
    group = parser.getgroup("katsdpsigproc")
    group.addoption(
        "--devices",
        choices=["first-per-api", "all", "none"],
        default="first-per-api",
        help="Select which devices to use for testing"
    )


def pytest_configure(config) -> None:
    config.addinivalue_line("markers", "force_autotune: unconditionally run autotuning")
    config.addinivalue_line("markers", "cuda_only: run test only on CUDA devices")
    config.addinivalue_line("markers", "opencl_only: run test only on OpenCL devices")
    config.addinivalue_line(
        "markers", "device_filter(filter): run test only on devices matching 'filter'")


def pytest_generate_tests(metafunc) -> None:
    if "device" in metafunc.fixturenames:
        option = metafunc.config.getoption("devices")
        if option == "none":
            metafunc.parametrize(
                "device",
                [
                    pytest.param(
                        None, marks=pytest.mark.skip(reason="--devices=none passed on command line")
                    )
                ]
            )
            return

        devices = accel.candidate_devices()
        # Apply filters. This is done after calling candidate_devices rather
        # than by passing a device filter, so that
        # environment variables like KATSDPSIGPROC_DEVICE are interpreted
        # relative to the full device list rather than a filtered list, which
        # could cause tests with different filters to select different
        # devices.
        for marker in metafunc.definition.iter_markers("cuda_only"):
            min_cc = marker.kwargs.get("min_compute_capability", (0, 0))
            devices = [
                device
                for device in devices
                if device.is_cuda and device.compute_capability >= min_cc]  # type: ignore
        if metafunc.definition.get_closest_marker("opencl_only") is not None:
            devices = [device for device in devices if not device.is_cuda]
        for marker in metafunc.definition.iter_markers("device_filter"):
            devices = [device for device in devices if marker.args[0](device)]

        # Apply --devices command-line setting
        if option == "first-per-api":
            # Sort so that itertools.groupby finds all devices from the same
            # API together.
            def classify_api(device: abc.AbstractDevice) -> bool:
                return device.is_cuda  # Currently only 2 APIs, so this is sufficient

            devices = sorted(devices, key=classify_api)
            # group is an iterable, so next(group) gives the first element
            devices = [next(group) for _, group in itertools.groupby(devices, key=classify_api)]
        # Nothing needed for --devices=all, and --devices=none handled earlier

        ids = [f"{d.name} ({d.platform_name})" for d in devices]
        if not devices:
            metafunc.parametrize(
                "device",
                [
                    pytest.param(
                        None, marks=pytest.mark.xfail(reason="No matching device found", run=False)
                    )
                ]
            )
        else:
            metafunc.parametrize("device", devices, ids=ids)



################################################################################
# Copyright (c) 2014-2021, National Research Foundation (SARAO)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

"""Transpose 2D arrays on a device."""

from typing import Tuple, Optional, Mapping, Callable, Any, cast
from typing_extensions import TypedDict

import numpy as np
try:
    from numpy.typing import DTypeLike
except ImportError:
    DTypeLike = Any     # type: ignore

from . import accel
from . import tune
from .abc import AbstractContext, AbstractCommandQueue


_TuningDict = TypedDict('_TuningDict', {'block': int, 'vtx': int, 'vty': int})


class TransposeTemplate:
    """Kernel for transposing a 2D array of data.

    Parameters
    ----------
    context
        Context for which kernels will be compiled
    dtype
        Type of data elements
    ctype
        Type (in C/CUDA, not numpy) of data elements
    tuning
        Kernel tuning parameters; if omitted, will autotune. The possible
        parameters are

        - block: number of workitems per workgroup in each dimension
        - vtx, vty: elements per workitem in each dimension
    """

    autotune_version = 1

    def __init__(self, context: AbstractContext, dtype: DTypeLike, ctype: str,
                 tuning: Optional[_TuningDict] = None) -> None:
        self.context = context
        self.dtype: np.dtype = np.dtype(dtype)
        self.ctype = ctype
        if tuning is None:
            tuning = self.autotune(context, dtype, ctype)
        self._block = tuning['block']
        self._tilex = tuning['block'] * tuning['vtx']
        self._tiley = tuning['block'] * tuning['vty']
        self.program = accel.build(context, "transpose.mako", {
            'block': tuning['block'],
            'vtx': tuning['vtx'],
            'vty': tuning['vty'],
            'ctype': ctype
        })

    @classmethod
    @tune.autotuner(test={'block': 8, 'vtx': 2, 'vty': 3})
    def autotune(cls, context: AbstractContext, dtype: DTypeLike, ctype: str) -> _TuningDict:
        queue = context.create_tuning_command_queue()
        in_shape = (2048, 2048)
        out_shape = (2048, 2048)
        in_data = accel.DeviceArray(context, in_shape, dtype=dtype)
        out_data = accel.DeviceArray(context, out_shape, dtype=dtype)

        def generate(block: int, vtx: int, vty: int) -> Optional[Callable[[int], float]]:
            local_mem = (block * vtx + 1) * (block * vty) * np.dtype(dtype).itemsize
            if local_mem > 32768:
                # Skip configurations using lots of lmem
                raise RuntimeError('too much local memory')
            fn = cls(context, dtype, ctype, {
                'block': block,
                'vtx': vtx,
                'vty': vty}).instantiate(queue, in_shape)
            fn.bind(src=in_data, dest=out_data)
            return tune.make_measure(queue, fn)

        return cast(_TuningDict, tune.autotune(generate,
                                               block=[4, 8, 16, 32],
                                               vtx=[1, 2, 3, 4],
                                               vty=[1, 2, 3, 4]))

    def instantiate(self, command_queue: AbstractCommandQueue, shape: Tuple[int, int],
                    allocator: Optional[accel.AbstractAllocator] = None) -> 'Transpose':
        return Transpose(self, command_queue, shape, allocator)


class Transpose(accel.Operation):
    """Concrete instance of :class:`TransposeTemplate`.

    .. rubric:: Slots

    **src**
        Input

    **dest**
        Output
    """

    def __init__(self, template: TransposeTemplate, command_queue: AbstractCommandQueue,
                 shape: Tuple[int, int], allocator: Optional[accel.AbstractAllocator] = None):
        super().__init__(command_queue, allocator)
        self.template = template
        self.kernel = template.program.get_kernel("transpose")
        self.shape = shape
        self.slots['src'] = accel.IOSlot(shape, template.dtype)
        self.slots['dest'] = accel.IOSlot((shape[1], shape[0]), template.dtype)

    def _run(self) -> None:
        src = self.buffer('src')
        dest = self.buffer('dest')
        # Round up to number of blocks in each dimension
        in_row_tiles = accel.divup(src.shape[0], self.template._tiley)
        in_col_tiles = accel.divup(src.shape[1], self.template._tilex)
        self.command_queue.enqueue_kernel(
            self.kernel,
            [
                dest.buffer, src.buffer,
                np.int32(src.shape[0]), np.int32(src.shape[1]),
                np.int32(dest.padded_shape[1]),
                np.int32(src.padded_shape[1])
            ],
            global_size=(in_col_tiles * self.template._block, in_row_tiles * self.template._block),
            local_size=(self.template._block, self.template._block))

    def parameters(self) -> Mapping[str, Any]:
        return {
            'dtype': self.template.dtype,
            'ctype': self.template.ctype,
            'shape': self.slots['src'].shape      # type: ignore
        }



################################################################################
# Copyright (c) 2021, 2025 National Research Foundation (SARAO)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

"""Fast Fourier Transforms.

Currently this module only supports CUDA, using cuFFT.

It does not use scikit-cuda, because that insists on creating a cublas context
to obtain a version number, which then permanently consumes GPU memory on some
arbitrary GPU. It also seems the maintainer `no longer has time
<https://github.com/lebedov/scikit-cuda/issues/319>`_.

Only Linux is supported. Windows and MacOS support would require changing the
code for locating the library.
"""

from enum import Enum
import ctypes.util
import threading
from typing import Any, Callable, Dict, Optional, Tuple

import numpy as np
try:
    from numpy.typing import DTypeLike
except ImportError:
    DTypeLike = Any  # type: ignore

from . import accel, cuda
from .accel import AbstractAllocator
from .abc import AbstractContext, AbstractCommandQueue


class FftMode(Enum):
    FORWARD = 0
    INVERSE = 1


class _CEnum(Enum):
    """Enum type that can be used with ctypes."""

    @classmethod
    def from_param(cls, value):
        if not isinstance(value, cls):
            raise TypeError
        # TODO: assumes that enum and int have the size in the ABI. But this
        # seems to be what scikit-cuda assumed anyway.
        return ctypes.c_int(value.value)


class _Cufft:
    """Wraps up the low-level ctypes access to cuFFT.

    This is not intended to be a complete wrapper for cuFFT, nor to be
    user-friendly. It provides just enough for the implementation of this
    module.
    """

    cufftHandle = ctypes.c_int
    cudaStream_t = ctypes.c_void_p  # It's a pointer to an opaque struct

    class cufftType(_CEnum):
        CUFFT_R2C = 0x2a
        CUFFT_C2R = 0x2c
        CUFFT_C2C = 0x29
        CUFFT_D2Z = 0x6a
        CUFFT_Z2D = 0x6c
        CUFFT_Z2Z = 0x69

    class cufftResult(_CEnum):
        CUFFT_SUCCESS = 0
        CUFFT_INVALID_PLAN = 1
        CUFFT_ALLOC_FAILED = 2
        CUFFT_INVALID_TYPE = 3
        CUFFT_INVALID_VALUE = 4
        CUFFT_INTERNAL_ERROR = 5
        CUFFT_EXEC_FAILED = 6
        CUFFT_SETUP_FAILED = 7
        CUFFT_INVALID_SIZE = 8
        CUFFT_UNALIGNED_DATA = 9
        CUFFT_INCOMPLETE_PARAMETER_LIST = 10
        CUFFT_INVALID_DEVICE = 11
        CUFFT_PARSE_ERROR = 12
        CUFFT_NO_WORKSPACE = 13
        CUFFT_NOT_IMPLEMENTED = 14
        CUFFT_LICENSE_ERROR = 15
        CUFFT_NOT_SUPPORTED = 16

    # cufft.h has no such type (these are just #defines), but it is convenient
    # to use an enum in Python.
    class cufftDirection(_CEnum):
        CUFFT_FORWARD = -1
        CUFFT_INVERSE = 1

    class CufftError(Exception):
        """Error raised from cuFFT"""

        def __init__(self, error_code: int):
            self.error_code = error_code
            try:
                name = _Cufft.cufftResult(error_code).name
            except ValueError:
                name = f"cuFFT error {error_code:#x}"
            super().__init__(name)

    @staticmethod
    def _errcheck(result, func, args) -> None:
        """Turn cufftResult into an exception."""
        if result:
            raise _Cufft.CufftError(result)

    def _get_function(self, name: str, argtypes: list) -> Callable[..., None]:
        """Shortcut to build a ctypes function wrapper.

        The function must return a :class:`cufftResult`.
        """
        func = getattr(self.lib, name)
        func.argtypes = argtypes
        func.restype = ctypes.c_int
        func.errcheck = self._errcheck
        return func

    def __init__(self):
        self.lib = ctypes.cdll.LoadLibrary(ctypes.util.find_library("cufft"))
        self.cufftGetVersion = self._get_function("cufftGetVersion", [ctypes.POINTER(ctypes.c_int)])
        self.cufftCreate = self._get_function("cufftCreate", [ctypes.POINTER(self.cufftHandle)])
        self.cufftDestroy = self._get_function("cufftDestroy", [self.cufftHandle])
        self.cufftMakePlanMany64 = self._get_function(
            "cufftMakePlanMany64",
            [
                self.cufftHandle,
                ctypes.c_int,                           # rank
                ctypes.POINTER(ctypes.c_longlong),      # n
                ctypes.POINTER(ctypes.c_longlong),      # inembed
                ctypes.c_longlong,                      # istride,
                ctypes.c_longlong,                      # idist
                ctypes.POINTER(ctypes.c_longlong),      # onembed
                ctypes.c_longlong,                      # ostride
                ctypes.c_longlong,                      # odist
                self.cufftType,                         # type
                ctypes.c_longlong,                      # batch
                ctypes.POINTER(ctypes.c_size_t)         # workSize
            ]
        )
        self.cufftSetAutoAllocation = self._get_function(
            "cufftSetAutoAllocation", [self.cufftHandle, ctypes.c_int]
        )
        self.cufftSetStream = self._get_function(
            "cufftSetStream", [self.cufftHandle, self.cudaStream_t]
        )
        self.cufftSetWorkArea = self._get_function(
            "cufftSetWorkArea", [self.cufftHandle, ctypes.c_void_p]
        )

        # The actual function signatures have typed pointers rather than void*, but
        # specifying that just making invocation more complicated, and requires
        # defining types like cufftComplex.
        self.cufftExec = {
            self.cufftType.CUFFT_R2C: self._get_function(
                "cufftExecR2C", [self.cufftHandle, ctypes.c_void_p, ctypes.c_void_p]
            ),
            self.cufftType.CUFFT_C2R: self._get_function(
                "cufftExecC2R", [self.cufftHandle, ctypes.c_void_p, ctypes.c_void_p]
            ),
            self.cufftType.CUFFT_D2Z: self._get_function(
                "cufftExecD2Z", [self.cufftHandle, ctypes.c_void_p, ctypes.c_void_p]
            ),
            self.cufftType.CUFFT_Z2D: self._get_function(
                "cufftExecZ2D", [self.cufftHandle, ctypes.c_void_p, ctypes.c_void_p]
            ),
            self.cufftType.CUFFT_C2C: self._get_function(
                "cufftExecC2C",
                [self.cufftHandle, ctypes.c_void_p, ctypes.c_void_p, self.cufftDirection]
            ),
            self.cufftType.CUFFT_Z2Z: self._get_function(
                "cufftExecZ2Z",
                [self.cufftHandle, ctypes.c_void_p, ctypes.c_void_p, self.cufftDirection]
            )
        }


class FftTemplate:
    r"""Operation template for a forward or reverse FFT.

    The transformation is done over the last N dimensions, with the remaining
    dimensions for batching multiple arrays to be transformed. Dimensions
    before the first N must not be padded.

    This template bakes in more information than most (data shapes), which is
    due to constraints in CUFFT.

    The template can specify real->complex, complex->real, or
    complex->complex. In the last case, the same template can be used to
    instantiate forward or inverse transforms. Otherwise, real->complex
    transforms must be forward, and complex->real transforms must be inverse.

    For real<->complex transforms, the final dimension of the padded shape
    need only be :math:`\lfloor\frac{L}{2}\rfloor + 1`, where :math:`L` is the
    last element of `shape`.

    The transform is unnormalised: performing a forward followed by a reverse
    transform will scale the result by the number of elements.

    Parameters
    ----------
    context
        Context for the operation
    N
        Number of dimensions for the transform
    shape
        Shape of the data (N or more dimensions). For real->complex or
        complex->real transformation, this is the size of the real side of the
        transform.
    dtype_src : {`np.float32`, `np.float64`, `np.complex64`, `np.complex128`}
        Data type for input
    dtype_dest : {`np.float32`, `np.float64`, `np.complex64`, `np.complex128`}
        Data type for output
    padded_shape_src
        Padded shape of the input
    padded_shape_dest
        Padded shape of the output
    tuning
        Tuning parameters (currently unused)
    """

    def __init__(
        self,
        context: AbstractContext,
        N: int,
        shape: Tuple[int, ...],
        dtype_src: DTypeLike,
        dtype_dest: DTypeLike,
        padded_shape_src: Tuple[int, ...],
        padded_shape_dest: Tuple[int, ...],
        tuning: Optional[Dict[str, Any]] = None
    ) -> None:
        if not isinstance(context, cuda.Context):
            raise TypeError('Only CUDA contexts are supported')
        if len(padded_shape_src) != len(shape):
            raise ValueError('padded_shape_src and shape must have same length')
        if len(padded_shape_dest) != len(shape):
            raise ValueError('padded_shape_dest and shape must have same length')
        if padded_shape_src[:-N] != shape[:-N]:
            raise ValueError('Source must not be padded on batch dimensions')
        if padded_shape_dest[:-N] != shape[:-N]:
            raise ValueError('Destination must not be padded on batch dimensions')

        if dtype_src == np.float32 and dtype_dest == np.complex64:
            fft_type = _Cufft.cufftType.CUFFT_R2C
        elif dtype_src == np.complex64 and dtype_dest == np.float32:
            fft_type = _Cufft.cufftType.CUFFT_C2R
        elif dtype_src == np.complex64 and dtype_dest == np.complex64:
            fft_type = _Cufft.cufftType.CUFFT_C2C
        elif dtype_src == np.float64 and dtype_dest == np.complex128:
            fft_type = _Cufft.cufftType.CUFFT_D2Z
        elif dtype_src == np.complex128 and dtype_dest == np.float64:
            fft_type = _Cufft.cufftType.CUFFT_Z2D
        elif dtype_src == np.complex128 and dtype_dest == np.complex128:
            fft_type = _Cufft.cufftType.CUFFT_Z2Z
        else:
            raise ValueError("Invalid combination of dtypes")

        cufft = _Cufft()
        self._cufft = cufft
        self.context: cuda.Context = context
        self.shape = shape
        # TODO: the type annotation is necessary with numpy 1.20.1, but
        # is no longer needed for newer versions.
        self.dtype_src: np.dtype = np.dtype(dtype_src)
        self.dtype_dest: np.dtype = np.dtype(dtype_dest)
        self.padded_shape_src = padded_shape_src
        self.padded_shape_dest = padded_shape_dest
        # CUDA 7.0 CUFFT has a bug where kernels are run in the default stream
        # instead of the requested one, if dimensions are up to 1920. There is
        # a patch, but there is no query to detect whether it has been
        # applied.
        version = ctypes.c_int()
        cufft.cufftGetVersion(ctypes.byref(version))
        self._needs_synchronize_workaround = (
            version.value < 8000 and any(x <= 1920 for x in shape[:N])
        )
        batches = int(np.prod(padded_shape_src[:-N]))
        with context:
            arr_type = ctypes.c_longlong * N
            work_size = ctypes.c_size_t()
            self._plan = cufft.cufftHandle()
            cufft.cufftCreate(ctypes.byref(self._plan))
            cufft.cufftSetAutoAllocation(self._plan, False)
            cufft.cufftMakePlanMany64(
                self._plan,                               # handle
                N,                                        # rank
                arr_type(*shape[-N:]),                    # n
                arr_type(*padded_shape_src[-N:]),         # inembed
                1,                                        # istride
                int(np.prod(padded_shape_src[-N:])),      # idist
                arr_type(*padded_shape_dest[-N:]),        # onembed
                1,                                        # ostride
                int(np.prod(padded_shape_dest[-N:])),     # odist
                fft_type,                                 # type
                batches,                                  # batch
                ctypes.byref(work_size)                   # workSize
            )
        self._work_size = work_size.value
        self._cufft = cufft
        self._fft_type = fft_type
        # The stream and work area are associated with the plan rather than
        # the execution, so we need to serialise executions.
        self._lock = threading.RLock()

    def instantiate(self, *args, **kwargs):
        return Fft(self, *args, **kwargs)

    def __del__(self):
        if hasattr(self, '_plan'):
            with self.context:
                self._cufft.cufftDestroy(self._plan)


class Fft(accel.Operation):
    """Forward or inverse Fourier transformation.

    .. rubric:: Slots

    **src**
        Input data
    **dest**
        Output data
    **work_area**
        Scratch area for work. The contents should not be used; it is made
        available so that it can be aliased with other scratch areas.
        If the implementation does not need any device memory for scratch
        space, this slot will not exist.

    Parameters
    ----------
    template
        Operation template
    command_queue
        Command queue for the operation
    mode
        FFT direction
    allocator
        Allocator used to allocate unbound slots
    """

    command_queue: cuda.CommandQueue  # refine the type for mypy

    def __init__(
        self,
        template: FftTemplate,
        command_queue: AbstractCommandQueue,
        mode: FftMode,
        allocator: Optional[AbstractAllocator] = None
    ) -> None:
        if not isinstance(command_queue, cuda.CommandQueue):
            raise TypeError('Only CUDA command queues are supported')
        super().__init__(command_queue, allocator)
        self.template = template
        src_shape = list(template.shape)
        dest_shape = list(template.shape)
        if template.dtype_src.kind != 'c':
            if mode != FftMode.FORWARD:
                raise ValueError('R2C transform must use FftMode.FORWARD')
            dest_shape[-1] = template.shape[-1] // 2 + 1
        if template.dtype_dest.kind != 'c':
            if mode != FftMode.INVERSE:
                raise ValueError('C2R transform must use FftMode.INVERSE')
            src_shape[-1] = template.shape[-1] // 2 + 1
        src_dims = tuple(accel.Dimension(d[0], min_padded_size=d[1], exact=True)
                         for d in zip(src_shape, template.padded_shape_src))
        dest_dims = tuple(accel.Dimension(d[0], min_padded_size=d[1], exact=True)
                          for d in zip(dest_shape, template.padded_shape_dest))
        self.slots['src'] = accel.IOSlot(src_dims, template.dtype_src)
        self.slots['dest'] = accel.IOSlot(dest_dims, template.dtype_dest)
        if template._work_size > 0:
            self.slots['work_area'] = accel.IOSlot((template._work_size,), np.uint8)
        self.mode = mode

    def _run(self) -> None:
        src_buffer = self.buffer('src')
        dest_buffer = self.buffer('dest')
        context = self.command_queue.context
        cufft = self.template._cufft
        with context, self.template._lock:
            cufft.cufftSetStream(
                self.template._plan,
                self.command_queue._pycuda_stream.handle
            )
            if "work_area" in self.slots:
                work_area_buffer = self.buffer('work_area')
                cufft.cufftSetWorkArea(
                    self.template._plan,
                    work_area_buffer.buffer.ptr
                )
            func = cufft.cufftExec[self.template._fft_type]
            args = [self.template._plan, src_buffer.buffer.ptr, dest_buffer.buffer.ptr]
            if len(func.argtypes) == 4:
                args.append(_Cufft.cufftDirection['CUFFT_' + self.mode.name])
            func(*args)
            if self.template._needs_synchronize_workaround:
                context._pycuda_context.synchronize()



#!/usr/bin/env python

################################################################################
# Copyright (c) 2014-2021, National Research Foundation (SARAO)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

import time
import numpy as np
from katsdpsigproc import accel
from katsdpsigproc import percentile as perc5

context = accel.create_some_context(True)
queue = context.create_command_queue(profile=True)

data = np.abs(np.random.randn(4000, 5000)).astype(np.float32)

template = perc5.Percentile5Template(context, max_columns=5000)
perc = template.instantiate(queue, data.shape)  # type: ignore [arg-type]
perc.ensure_all_bound()
perc.buffer('src').set(queue, data)
start_event = queue.enqueue_marker()
perc()
end_event = queue.enqueue_marker()
out = perc.buffer('dest').get(queue)

t0 = time.time()
expected = np.percentile(data, [0, 100, 25, 75, 50], axis=1, interpolation='lower')
t1 = time.time()
print('gpu:', end_event.time_since(start_event), 'cpu:', t1 - t0)
np.testing.assert_equal(out, expected)



################################################################################
# Copyright (c) 2014-2021, National Research Foundation (SARAO)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

"""RFI flagging algorithms that run on the CPU."""

from abc import ABC, abstractmethod
from typing import Optional

import numpy as np
import pandas as pd

from . import MAD_NORMAL


class AbstractBackgroundHost(ABC):
    @abstractmethod
    def __init__(self, width: int, amplitudes: bool = False) -> None:
        pass    # pragma: nocover

    @abstractmethod
    def __call__(self, vis: np.ndarray, flags: Optional[np.ndarray] = None) -> np.ndarray:
        """Subtract an estimate of background signal (without RFI).

        Parameters
        ----------
        vis : array-like, float or complex
            Visibilities, either as complex values or amplitudes, depending on
            how the class was constructed.
        flags : array-like, int
            Initial flags, indicating data known to be bad prior to RFI flagging.

        Returns
        -------
        deviations
            Amplitude of signal after subtracting background.
        """


class AbstractNoiseEstHost(ABC):
    @abstractmethod
    def __call__(self, deviations: np.ndarray) -> np.ndarray:
        """Compute the estimated standard deviation of noise per baseline.

        Parameters
        ----------
        deviations
            Deviations from the background amplitude, indexed by channel
            then baseline.

        Returns
        -------
        noise
            1D array of `np.float32`, containing the noise estimate for each channel
        """


class AbstractThresholdHost(ABC):
    @abstractmethod
    def __init__(self, n_sigma: float) -> None:
        pass        # pragma: nocover

    @abstractmethod
    def __call__(self, deviations: np.ndarray, noise: np.ndarray) -> np.ndarray:
        """Apply the thresholding.

        Parameters
        ----------
        deviations : array-like, real
            Deviations from the background amplitude, indexed by channel
            then baseline.
        noise : array-like, real
            Per-baseline estimate of the standard deviation for noise

        Returns
        -------
        array-like
            Array of `np.uint8`, containing the flag value or 0
        """


class AbstractFlaggerHost(ABC):
    @abstractmethod
    def __call__(self, vis: np.ndarray, input_flags: Optional[np.ndarray] = None) -> np.ndarray:
        """Perform the flagging.

        Parameters
        ----------
        vis
            The input visibilities as a 2D array of complex64 (or float, if
            that's what the backgrounder expects), indexed by channel and
            baseline.
        input_flags
            Predefined flags as an array of uint8. These can be either
            a 1D array of per-channel flags or a 2D array with the same
            shape as `vis`.

        Returns
        -------
        flags
            Flags of the same shape as `vis`. Note that `input_flags`
            are not copied into the output.
        """


class BackgroundMedianFilterHost(AbstractBackgroundHost):
    """Host backgrounder that applies a median filter to each baseline (by amplitude).

    Parameters
    ----------
    width
        The kernel width (must be odd)
    amplitudes
        If `True`, the inputs are amplitudes rather than complex visibilities
    """

    def __init__(self, width: int, amplitudes: bool = False) -> None:
        self.width = width
        self.amplitudes = amplitudes

    def __call__(self, vis: np.ndarray, flags: Optional[np.ndarray] = None) -> np.ndarray:
        if self.amplitudes:
            amp = pd.DataFrame(vis)
        else:
            amp = pd.DataFrame(np.abs(vis))
        if flags is not None:
            # Convert the flags to bool, and add in a baseline axis if not
            # already present. The mask function in Pandas doesn't
            # automatically broadcast, so we have to do so explicitly with
            # np.broadcast_to.
            flags = flags.astype(np.bool_)
            if flags.ndim < 2:
                flags = flags[:, np.newaxis]
            flags_2d = np.broadcast_to(flags, vis.shape)
            amp = amp.mask(flags_2d)
        med = amp.rolling(self.width, center=True, min_periods=1).median()
        deviation = amp - med
        deviation.fillna(0, inplace=True)
        return deviation.values


class NoiseEstMADHost(AbstractNoiseEstHost):
    """Estimate noise using the median of non-zero absolute deviations."""

    def __call__(self, deviations: np.ndarray) -> np.ndarray:
        baselines = deviations.shape[1]
        out = np.empty(baselines)
        for i in range(baselines):
            abs_dev = np.abs(deviations[:, i])
            out[i] = np.median(abs_dev[abs_dev > 0])
        return out * MAD_NORMAL


class ThresholdSimpleHost(AbstractThresholdHost):
    """Threshold each element independently.

    Parameters
    ----------
    n_sigma
        Number of (estimated) standard deviations for the threshold
    flag_value
        Number stored in returned value to indicate RFI
    """

    def __init__(self, n_sigma: float, flag_value: int = 1) -> None:
        self.n_sigma = n_sigma
        self.flag_value = flag_value

    def __call__(self, deviations: np.ndarray, noise: np.ndarray) -> np.ndarray:
        flags = (deviations > self.n_sigma * noise).astype(np.uint8)
        return flags * self.flag_value


class ThresholdSumHost(AbstractThresholdHost):
    """Thresholding using the Offringa Sum-Threshold algorithm, with power-of-two sized windows.

    The initial (single-pixel) threshold is determined by median of absolute
    deviations.

    At present, auto- and cross-correlations are treated the same.

    Parameters
    ----------
    n_sigma
        Number of (estimated) standard deviations for the threshold
    n_windows
        Number of window sizes to use
    threshold_falloff
        Controls rate at which thresholds decrease (ρ in Offringa 2010)
    flag_value
        Number stored in returned value to indicate RFI
    """

    def __init__(self, n_sigma: float, n_windows: int = 4,
                 threshold_falloff: float = 1.2, flag_value: int = 1) -> None:
        self.n_sigma = n_sigma
        self.windows = [2 ** i for i in range(n_windows)]
        self.threshold_scales = [pow(threshold_falloff, -i) for i in range(n_windows)]
        self.flag_value = flag_value

    def apply_baseline(self, deviations: np.ndarray, threshold1: float) -> np.ndarray:
        """Apply the thresholding to a single baseline.

        The flags are returned as booleans, rather than
        `flag_value`.

        Parameters
        ----------
        deviations : 1D array, float
            Deviations from the background
        threshold1
            Threshold for RFI on individual samples
        """
        # The data are modified, so use a copy
        deviations = deviations.copy()
        flags = np.zeros_like(deviations, dtype=np.bool_)
        for window, scale in zip(self.windows, self.threshold_scales):
            threshold = np.float32(threshold1 * scale)
            # Force already identified outliers to the threshold
            deviations[flags] = threshold
            # Compute sums
            weight = np.ones(window)
            sums = np.convolve(deviations, weight, mode='valid')
            # Identify outlier sums
            sum_flags = sums > threshold * window
            # Distribute flags
            weight = np.ones(window, dtype=np.bool_)
            flags |= np.convolve(sum_flags, weight)
        return flags

    def __call__(self, deviations: np.ndarray, noise: np.ndarray) -> np.ndarray:
        flags = np.empty_like(deviations, dtype=np.uint8)
        baselines = deviations.shape[1]
        for i in range(baselines):
            bl_flags = self.apply_baseline(deviations[:, i], self.n_sigma * noise[i])
            flags[:, i] = bl_flags * np.uint8(self.flag_value)
        return flags


class FlaggerHost(AbstractFlaggerHost):
    """Combine host background and thresholding implementations to make a flagger."""

    def __init__(self, background: AbstractBackgroundHost, noise_est: AbstractNoiseEstHost,
                 threshold: AbstractThresholdHost):
        self.background = background
        self.noise_est = noise_est
        self.threshold = threshold

    def __call__(self, vis: np.ndarray, input_flags: Optional[np.ndarray] = None) -> np.ndarray:
        deviations = self.background(vis, input_flags)
        noise = self.noise_est(deviations)
        return self.threshold(deviations, noise)



################################################################################
# Copyright (c) 2014-2022, National Research Foundation (SARAO)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

"""Tests for RFI noise estimation algorithms."""

from abc import ABC, abstractmethod

import numpy as np
import pytest

from katsdpsigproc.rfi import host, device
from katsdpsigproc.abc import AbstractContext, AbstractCommandQueue


_deviations: np.ndarray
_deviations_big: np.ndarray
_expected: np.ndarray


def setup():   # type: () -> None
    global _deviations, _deviations_big, _expected
    _deviations = np.array(
        [
            [0.0, 3.0, 2.4],
            [1.5, -1.4, 4.6],
            [0.0, 1.1, 3.3],
            [5.0, 0.0, -3.1]
        ]).astype(np.float32)
    _expected = np.array([3.25, 1.4, 3.2]) * 1.4826
    shape = (117, 273)

    # Use a fixed seed to make the test repeatable
    rs = np.random.RandomState(seed=1)
    _deviations_big = rs.standard_normal(shape).astype(np.float32)


def test_NoiseEstMADHost() -> None:
    global _deviations, _expected
    noise_est = host.NoiseEstMADHost()
    actual = noise_est(_deviations)
    np.testing.assert_allclose(_expected, actual)


class BaseTestNoiseEstDeviceClass(ABC):
    def test_result(self, context: AbstractContext, command_queue: AbstractCommandQueue) -> None:
        global _deviations_big
        template = self.factory(context)
        ne_host = template.host_class()
        ne_device = device.NoiseEstHostFromDevice(template, command_queue)
        noise_host = ne_host(_deviations_big)
        noise_device = ne_device(_deviations_big)
        np.testing.assert_allclose(noise_host, noise_device)

    @pytest.mark.force_autotune
    def test_autotune(self, context: AbstractContext, command_queue: AbstractCommandQueue) -> None:
        self.factory(context)

    @abstractmethod
    def factory(self, context: AbstractContext) -> device.AbstractNoiseEstDeviceTemplate:
        pass       # pragma: nocover


class TestNoiseEstMADDevice(BaseTestNoiseEstDeviceClass):
    def factory(self, context: AbstractContext) -> device.NoiseEstMADDeviceTemplate:
        return device.NoiseEstMADDeviceTemplate(context)


class TestNoiseEstMADTDevice(BaseTestNoiseEstDeviceClass):
    def factory(self, context: AbstractContext) -> device.NoiseEstMADTDeviceTemplate:
        return device.NoiseEstMADTDeviceTemplate(context, 10240)



################################################################################
# Copyright (c) 2019-2021, National Research Foundation (SARAO)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

"""Abstract base classes for :mod:`.opencl` and :mod:`.cuda`."""

from abc import ABC, abstractmethod
from typing import List, Tuple, Sequence, Optional, Any, Type, TypeVar, Generic
from types import TracebackType

import numpy as np
try:
    from numpy.typing import DTypeLike
except ImportError:
    DTypeLike = Any     # type: ignore


_B = TypeVar('_B')     # buffer type
_RB = TypeVar('_RB')   # raw buffer type
_RS = TypeVar('_RS')   # raw buffer for SVM
_C = TypeVar('_C', bound='AbstractContext')
_D = TypeVar('_D', bound='AbstractDevice')
_E = TypeVar('_E', bound='AbstractEvent')
_K = TypeVar('_K', bound='AbstractKernel')
_P = TypeVar('_P', bound='AbstractProgram')
_Q = TypeVar('_Q', bound='AbstractCommandQueue')
_TQ = TypeVar('_TQ', bound='AbstractTuningCommandQueue')


class AbstractProgram(ABC, Generic[_K]):
    """Abstraction of a program object."""

    @abstractmethod
    def get_kernel(self, name: str) -> 'AbstractKernel':
        """Create a new kernel.

        Parameters
        ----------
        name
            Name of the kernel function
        """


class AbstractKernel(ABC, Generic[_P]):
    """Abstraction of a kernel object.

    The object can be enqueued using :meth:`AbstractCommandQueue.enqueue_kernel`.

    The recommended way to create this object is via
    :meth:`AbstractProgram.get_kernel`.
    """

    @abstractmethod
    def __init__(self, program: _P, name: str) -> None:
        pass


class AbstractEvent(ABC):
    """Abstraction of an event.

    This is more akin to a CUDA event than an OpenCL event, in that it is a
    marker in a command queue rather than associated with a specific command.
    """

    @abstractmethod
    def wait(self) -> None:
        """Block until the event has completed."""

    @abstractmethod
    def time_since(self: _E, prior_event: _E) -> float:
        """Return the time in seconds from `prior_event` to self.

        Unlike the PyCUDA method of the same name, this will wait for the
        events to complete if they have not already.
        """

    @abstractmethod
    def time_till(self: _E, next_event: _E) -> float:
        """Return the time in seconds from this event to `next_event`.

        See  :meth:`time_since`.
        """


class AbstractDevice(ABC, Generic[_C]):
    """Abstraction of a device."""

    @abstractmethod
    def make_context(self) -> 'AbstractContext':
        """Create a new context associated with this device."""

    @property
    @abstractmethod
    def name(self) -> str:
        """Return human-readable name for the device."""

    @property
    @abstractmethod
    def platform_name(self) -> str:
        """Return human-readable name for the platform owning the device."""

    @property
    @abstractmethod
    def driver_version(self) -> str:
        """Return human-readable name for the driver version."""

    @property
    @abstractmethod
    def is_cuda(self) -> bool:
        """Whether the device is a CUDA device."""

    @property
    @abstractmethod
    def is_gpu(self) -> bool:
        """Whether device is a GPU."""

    @property
    @abstractmethod
    def is_accelerator(self) -> bool:
        """Whether device is an accelerator (as defined by OpenCL device types)."""

    @property
    @abstractmethod
    def is_cpu(self) -> bool:
        """Whether the device is a CPU."""

    @property
    @abstractmethod
    def simd_group_size(self) -> int:
        """Return the number of workitems that run in lock-step.

        This must only be used to tune performance parameters; there are no
        guarantees about memory coherency, forward progress etc.
        """

    @classmethod
    @abstractmethod
    def get_devices(cls: Type[_D]) -> Sequence[_D]:
        """Return a list of all devices on all platforms."""

    @classmethod
    @abstractmethod
    def get_devices_by_platform(cls: Type[_D]) -> Sequence[Sequence[_D]]:
        """Return a list of all devices, with a sub-list per platform."""


class AbstractContext(ABC, Generic[_B, _RB, _RS, _D, _P, _Q, _TQ]):
    """Abstraction of an OpenCL/CUDA context."""

    @property
    @abstractmethod
    def device(self) -> AbstractDevice:
        """Return the device associated with the context (or the first device, if multiple)."""

    @abstractmethod
    def compile(self, source: str, extra_flags: Optional[List[str]] = None) -> AbstractProgram:
        """Build a program object from source.

        Parameters
        ----------
        source
            Source code
        extra_flags
            Extra parameters to pass to the compiler
        """

    @abstractmethod
    def allocate_raw(self, n_bytes: int) -> _RB:
        """Create an untyped buffer on the device."""

    @abstractmethod
    def allocate(self, shape: Tuple[int, ...], dtype: DTypeLike, raw: Optional[_RB] = None) -> _B:
        """Create a typed buffer on the device.

        Parameters
        ----------
        shape
            Shape for the array
        dtype
            Type for the data
        raw
            Memory backing the array (automatically allocated if ``None``)
        """

    @abstractmethod
    def allocate_pinned(self, shape: Tuple[int, ...], dtype: DTypeLike) -> np.ndarray:
        """Create a buffer in host memory that can be efficiently copied to and from the device.

        Parameters
        ----------
        shape
            Shape for the array
        dtype
            Type for the data
        """

    @abstractmethod
    def allocate_svm_raw(self, n_bytes: int) -> _RS:
        """Allocate raw storage that can be passed to :meth:`allocate_svm`."""

    @abstractmethod
    def allocate_svm(self, shape: Tuple[int, ...], dtype: DTypeLike,
                     raw: Optional[_RS] = None) -> np.ndarray:
        """Allocate shared virtual memory."""

    @abstractmethod
    def create_command_queue(self, profile: bool = False) -> 'AbstractCommandQueue':
        """Create a new command queue associated with this context.

        Parameters
        ----------
        profile
            If true, the command queue will support timing kernels
        """

    @abstractmethod
    def create_tuning_command_queue(self) -> 'AbstractTuningCommandQueue':
        """Create a new command queue for doing autotuning."""

    @abstractmethod
    def __enter__(self: _C) -> _C:
        pass

    @abstractmethod
    def __exit__(self,
                 exc_type: Optional[Type[BaseException]],
                 exc_val: Optional[BaseException],
                 exc_tb: Optional[TracebackType]) -> None:
        pass


class AbstractCommandQueue(ABC, Generic[_B, _C, _E, _K]):
    """Abstraction of a command queue."""

    context = None     # type: _C

    @abstractmethod
    def enqueue_read_buffer(self, buffer: _B, data: Any, blocking: bool = True) -> None:
        """Copy data from the device to the host.

        Only whole-buffer copies are supported, and the shape and type must
        match. In general, one should use the convenience functions in
        :class:`accel.DeviceArray`.

        Parameters
        ----------
        buffer
            Source
        data
            Target
        blocking
            If true (default) the call blocks until the copy is complete
        """

    @abstractmethod
    def enqueue_write_buffer(self, buffer: _B, data: Any, blocking=True) -> None:
        """Copy data from the host to the device.

        Only whole-buffer copies are supported, and the shape and type must
        match. In general, one should use the convenience functions in
        :class:`accel.DeviceArray`.

        Parameters
        ----------
        buffer
            Target
        data : array-like
            Source
        blocking
            If true (default), the call blocks until the source has been fully
            read (it has not necessarily reached the device).
        """

    @abstractmethod
    def enqueue_copy_buffer_rect(
            self, src_buffer: _B, dest_buffer: _B, src_origin: int, dest_origin: int,
            shape: Sequence[int], src_strides: Sequence[int], dest_strides: Sequence[int]) -> None:
        """Copy a subregion of one buffer to another.

        This is a low-level interface that ignores the shape, strides etc of
        the buffers, and treats them as byte arrays. It also only supports 3 or
        fewer dimensions. Use
        :py:meth:`~katsdpsigproc.accel.DeviceArray.copy_region` for a
        high-level interface.

        Parameters
        ----------
        src_buffer,dest_buffer
            Source and destination buffers
        src_origin,dest_origin
            Offsets for the start of the copy, in bytes
        shape
            Shape of the region to copy (1-3 elements). The first dimension is
            a byte count.
        src_strides,dest_strides
            Strides for the source and destination memory layout, with the same
            length as `shape`. The first element of each must be 1, and each
            element must be a factor of the next element.
        """

    @abstractmethod
    def enqueue_read_buffer_rect(
            self, buffer: _B, data: Any,
            buffer_origin: int, data_origin: int, shape: Sequence[int],
            buffer_strides: Sequence[int], data_strides: Sequence[int],
            blocking: bool = True) -> None:
        """Copy a region of a buffer to host memory.

        This is a low-level interface that ignores the shape, strides etc of
        the buffers, and treats them as byte arrays. It also only supports 3 or
        fewer dimensions. Use
        :py:meth:`~katsdpsigproc.accel.DeviceArray.set_region` for a high-level
        interface.

        Parameters
        ----------
        buffer
            Source
        data : array-like
            Target
        buffer_origin, data_origin
            Offsets for the start of the copy, in bytes
        shape
            Shape of the region to copy (1-3 elements). The first dimension is
            a byte count.
        buffer_strides,data_strides
            Strides for the destination and source memory layout, with the same
            length as `shape`. The first element of each must be 1, and each
            element must be a factor of the next element.
        blocking
            If true, block until the transfer is complete.
        """

    @abstractmethod
    def enqueue_write_buffer_rect(
            self, buffer: _B, data: Any,
            buffer_origin: int, data_origin: int, shape: Sequence[int],
            buffer_strides: Sequence[int], data_strides: Sequence[int],
            blocking: bool = True) -> None:
        """Copy a region of host memory to a buffer.

        This is a low-level interface that ignores the shape, strides etc of
        the buffers, and treats them as byte arrays. It also only supports 3 or
        fewer dimensions. Use
        :py:meth:`~katsdpsigproc.accel.DeviceArray.set_region` for a high-level
        interface.

        Parameters
        ----------
        buffer
            Target
        data : array-like
            Source
        buffer_origin, data_origin
            Offsets for the start of the copy, in bytes
        shape
            Shape of the region to copy (1-3 elements). The first dimension is
            a byte count.
        buffer_strides,data_strides
            Strides for the destination and source memory layout, with the same
            length as `shape`. The first element of each must be 1, and each
            element must be a factor of the next element.
        blocking
            If true, block until the transfer is complete.
        """

    @abstractmethod
    def enqueue_zero_buffer(self, buffer: _B) -> None:
        """Fill a buffer with zero bytes."""

    @abstractmethod
    def enqueue_kernel(self, kernel: _K, args: Sequence[Any],
                       global_size: Tuple[int, ...], local_size: Tuple[int, ...]) -> None:
        """Enqueue a kernel to the command queue.

        .. warning:: It is not thread-safe to call this function in two threads
            on the same kernel at the same time.

        Parameters
        ----------
        kernel
            Kernel to run
        args
            Arguments to pass to the kernel. Refer to the PyOpenCL/CUDA
            documentation for details. Additionally, this function allows
            a low-level device array to be passed.
        global_size
            Number of work-items in each global dimension
        local_size
            Number of work-items in each local dimension. Must divide
            exactly into `global_size`.
        """

    @abstractmethod
    def enqueue_marker(self) -> AbstractEvent:
        """Create an event at this point in the command queue."""

    @abstractmethod
    def enqueue_wait_for_events(self, events: Sequence[_E]) -> None:
        """Enqueue a barrier to wait for all events in `events`."""

    @abstractmethod
    def flush(self) -> None:
        """Start enqueued work running, but do not wait for it to complete."""

    @abstractmethod
    def finish(self) -> None:
        """Block until all enqueued work has completed."""


class AbstractTuningCommandQueue(AbstractCommandQueue[_B, _C, _E, _K]):
    """Command queue with extra facilities for autotuning.

    It keeps track of kernels that are enqueued since the last call to
    :meth:`start_tuning`, and reports the total time they consume when
    :meth:`stop_tuning` is called.
    """

    @abstractmethod
    def start_tuning(self) -> None:
        pass

    @abstractmethod
    def stop_tuning(self) -> float:
        pass



#!/usr/bin/env python
import katsdpsigproc.accel

ctx = katsdpsigproc.accel.create_some_context(interactive=True)
print(f'Successfully created context on {ctx.device.name} ({ctx.device.platform_name})')



################################################################################
# Copyright (c) 2016-2022, National Research Foundation (SARAO)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

"""Utilities for scheduling device operations with asyncio."""

import asyncio
import logging
import collections
from typing import List, Deque, Awaitable, Iterable, Optional, Type, Generic, TypeVar
from types import TracebackType

from .abc import AbstractEvent


_T = TypeVar('_T')
_logger = logging.getLogger(__name__)


async def wait_until(future: Awaitable[_T], when: float,
                     loop: Optional[asyncio.AbstractEventLoop] = None) -> _T:
    """Like :func:`asyncio.wait_for`, but with an absolute timeout."""
    def ready(*args) -> None:
        if not waiter.done():
            waiter.set_result(None)

    if loop is None:
        loop = asyncio.get_event_loop()
    waiter = asyncio.Future(loop=loop)    # type: asyncio.Future[None]
    timeout_handle = loop.call_at(when, ready)
    # Ensure that the future is really a future, not a coroutine object
    future = asyncio.ensure_future(future, loop=loop)
    future.add_done_callback(ready)
    try:
        await waiter
        if future.done():
            return future.result()
        else:
            future.remove_done_callback(ready)
            future.cancel()
            raise asyncio.TimeoutError()
    finally:
        timeout_handle.cancel()


async def async_wait_for_events(events: Iterable[AbstractEvent],
                                loop: Optional[asyncio.AbstractEventLoop] = None) -> None:
    """Coroutine that waits for a list of device events."""
    def wait_for_events(events: List[AbstractEvent]) -> None:
        for event in events:
            event.wait()
        # Remove references to events before the future is resolved. This
        # prevents a race condition where a caller might await
        # async_wait_for_events, then drop its references to the events before
        # the executor's worker thread has a chance to, causing the event to be
        # destroyed in the worker thread. That in turn leads to a warning from
        # PyCUDA (and a resource leak) if the context can't be made current in
        # the worker thread.
        events.clear()

    if loop is None:
        loop = asyncio.get_event_loop()
    events = list(events)
    if events:
        await loop.run_in_executor(None, wait_for_events, events)


class ResourceAllocation(Generic[_T]):
    """A handle representing a future acquisition of a resource.

    There are two ways to make the acquisition current:

     1. Call :meth:`wait`, which returns a future. The result of this future
        is a list of device events that must complete before it is safe to use
        the resource; they can either be waited for on the host (for example,
        using ``run_in_executor``), or by enqueuing a device wait before
        device operations on the resource.
     2. The above steps (with a blocking host wait) can be combined using
        :meth:`wait_events`.

    Instances of this class should never be constructed directly. Instead, use
    :meth:`Resource.acquire`.

    This class implements the context manager protocol, providing the
    underlying resource as the return value. This handles some cleanup if the
    method using the resource raises an exception without releasing the
    resource cleanly.
    """

    def __init__(self,
                 start: 'asyncio.Future[List[AbstractEvent]]',
                 end: 'asyncio.Future[List[AbstractEvent]]',
                 value: _T, loop: asyncio.AbstractEventLoop) -> None:
        self._start = start
        self._end = end
        self._loop = loop
        self.value = value

    def wait(self) -> 'asyncio.Future[List[AbstractEvent]]':
        """Return a future that will be set to a list of device events to waited for."""
        return self._start

    async def wait_events(self) -> None:
        """Wait for previous use of the resource to be complete on the host.

        This is a coroutine.
        """
        events = await self._start
        await async_wait_for_events(events, loop=self._loop)

    def ready(self, events: Optional[List[AbstractEvent]] = None) -> None:
        """Indicate that we are done with the resource, and that subsequent acquirers may use it.

        Note that even if the caller decides that it doesn't need to use the
        resource, it must not call this until the resource is ready.

        Parameters
        ----------
        events
            Device events that must be waited on before the resource is ready
        """
        if events is None:
            events = []
        self._end.set_result(events)

    def __enter__(self) -> _T:
        return self.value

    def __exit__(self, exc_type: Optional[Type[BaseException]],
                 exc_value: Optional[BaseException],
                 exc_tb: Optional[TracebackType]) -> None:
        if not self._end.done():
            if exc_value is not None:
                self._end.set_exception(exc_value)
                # Prevent asyncio complaining that the exception was never
                # retrieved, because we're also propagating the exception
                # from the context manager block and it's not necessary for
                # the application to consume it a second time.
                self._end.exception()
            else:
                _logger.warning('Resource allocation was not explicitly made ready')
                self.ready()


class Resource(Generic[_T]):
    """Abstraction of a contended resource, which may exist on a device.

    Passing of ownership is done via futures. Acquiring a resource is a
    non-blocking operation that returns two futures: a future to wait for
    before use, and a future to be signalled with a result when done. The
    value of each of these futures is a (possibly empty) list of device
    events which must be waited on before more device work is scheduled.

    Parameters
    ----------
    value : object
        Underlying resource to manage
    loop
        Event loop used for asynchronous operations. If not specified, defaults
        to ``asyncio.get_event_loop()``.

    Attributes
    ----------
    value : object
        Underlying resource passed to the constructor
    """

    def __init__(self, value: _T, loop: Optional[asyncio.AbstractEventLoop] = None) -> None:
        if loop is None:
            loop = asyncio.get_event_loop()
        self._loop = loop
        self._future = asyncio.Future(loop=loop)  # type: asyncio.Future[List[AbstractEvent]]
        self._future.set_result([])
        self.value = value

    def acquire(self) -> ResourceAllocation[_T]:
        """Indicate intent to acquire the resource.

        This does not actually acquire the resource, but instead returns a
        handle that can be used to acquire and release it later. Acquisitions
        always occur in the order in which calls to :meth:`acquire` are made.

        See :class:`ResourceAllocation` for further details.
        """
        old = self._future
        self._future = asyncio.Future(loop=self._loop)
        return ResourceAllocation(old, self._future, self.value, loop=self._loop)


class JobQueue:
    """Maintain a list of in-flight asynchronous jobs."""

    def __init__(self) -> None:
        self._jobs = collections.deque()   # type: Deque[asyncio.Future]

    def add(self, job: Awaitable) -> None:
        """Append a job to the list.

        If `job` is a coroutine, it is automatically wrapped in a task.
        """
        self._jobs.append(asyncio.ensure_future(job))

    def clean(self) -> None:
        """Remove completed jobs from the front of the queue."""
        while self._jobs and self._jobs[0].done():
            self._jobs.popleft().result()     # Re-throws any exception

    async def finish(self, max_remaining: int = 0) -> None:
        """Wait for jobs to finish until there are at most `max_remaining` in the queue.

        This is a coroutine.
        """
        while len(self._jobs) > max_remaining:
            await self._jobs.popleft()

    def __len__(self) -> int:
        return len(self._jobs)

    def __bool__(self) -> bool:
        return bool(self._jobs)

    def __contains__(self, item: asyncio.Future) -> bool:
        return item in self._jobs


__all__ = ['wait_until', 'async_wait_for_events', 'Resource', 'ResourceAllocation', 'JobQueue']



################################################################################
# Copyright (c) 2014-2022, National Research Foundation (SARAO)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

"""Tests for RFI thresholding algorithms."""

from abc import ABC, abstractmethod
from typing import Type

import numpy as np
import pytest

from katsdpsigproc.rfi import host, device
from katsdpsigproc.abc import AbstractContext, AbstractCommandQueue


_deviations: np.ndarray
_spikes: np.ndarray


def setup():   # type: () -> None
    global _deviations, _spikes
    shape = (117, 273)
    # Use a fixed seed to make the test repeatable
    rs = np.random.RandomState(seed=1)
    # Pick 1/4 of samples to be RFI
    _spikes = rs.random_sample(shape) < 0.25
    _deviations = rs.standard_normal(shape).astype(np.float32) * 10.0
    _deviations[_spikes] += 200.0


def test_ThresholdSimpleHost() -> None:
    check_host_class(host.ThresholdSimpleHost, 11.0)


def test_ThresholdSumHost() -> None:
    check_host_class(host.ThresholdSumHost, 11.0)


def check_host_class(cls: Type[host.AbstractThresholdHost], n_sigma: float) -> None:
    global _deviations, _spikes
    threshold = cls(n_sigma)
    noise = np.repeat(10.0, _deviations.shape[1]).astype(np.float32)
    flags = threshold(_deviations, noise)
    np.testing.assert_equal(flags.astype(np.bool_), _spikes)


class BaseTestDeviceClass(ABC):
    def test_result(self, context: AbstractContext, command_queue: AbstractCommandQueue) -> None:
        global _deviations
        n_sigma = 11.0
        template = self.factory(context)
        th_host = template.host_class(n_sigma)
        th_device = device.ThresholdHostFromDevice(template, command_queue, n_sigma=n_sigma)
        noise = np.linspace(0.0, 50.0, _deviations.shape[1]).astype(np.float32)
        flags_host = th_host(_deviations, noise)
        flags_device = th_device(_deviations, noise)
        np.testing.assert_equal(flags_host, flags_device)

    @pytest.mark.force_autotune
    def test_autotune(self, context: AbstractContext, command_queue: AbstractCommandQueue) -> None:
        self.factory(context)

    @abstractmethod
    def factory(self, context: AbstractContext) -> device.AbstractThresholdDeviceTemplate:
        pass        # pragma: nocover


class TestThresholdSimpleDevice(BaseTestDeviceClass):
    def factory(self, context: AbstractContext) -> device.ThresholdSimpleDeviceTemplate:
        return device.ThresholdSimpleDeviceTemplate(context, False)


class TestThresholdSimpleDeviceTransposed(BaseTestDeviceClass):
    def factory(self, context: AbstractContext) -> device.ThresholdSimpleDeviceTemplate:
        return device.ThresholdSimpleDeviceTemplate(context, True)


class TestThresholdSumDevice(BaseTestDeviceClass):
    def factory(self, context: AbstractContext) -> device.ThresholdSumDeviceTemplate:
        return device.ThresholdSumDeviceTemplate(context)
